import os
import pickle
import tensorflow
import vtk
import glob
import numpy as np
from scipy.spatial import ConvexHull
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D




def visualiseMesh(vertices, triangles, colors):

    nodes = getPoints(vertices)
    elements = getElements(triangles)

    # Define the surface geometry
    polydata = vtk.vtkPolyData()
    # Specify that the variable 'nodes' as the points in the polydata
    polydata.SetPoints(nodes)
    # Similarly, specify that the variable 'elements' as the polys in polydata
    polydata.SetPolys(elements)

    # Assign the colors to the polydata
    assignColors(polydata, colors)

    # Create and run the visualization
    # ------------------------------

    # Create the renderer
    ren = vtk.vtkRenderer()
    # Create the window
    renWin = vtk.vtkRenderWindow()
    # Add the renderer to the window
    renWin.AddRenderer(ren)
    # Define an window interactor. This controls how the user can interact with
    # things in the window.
    iren = vtk.vtkRenderWindowInteractor()
    # Set the window defined above as the window that the interactor
    # will work on.
    iren.SetRenderWindow(renWin)

    # Initialize the mapper
    surfaceMapper = vtk.vtkPolyDataMapper()
    # Set the mapper's polydata as spherePolyData
    surfaceMapper.SetInputData(polydata)

    # Initialize the actor
    surfaceActor = vtk.vtkActor()
    # Set the actor's mapper as sphereMapper
    surfaceActor.SetMapper(surfaceMapper)
    # Add the actor to the renderer
    ren.AddActor(surfaceActor)
    #surfaceActor.GetProperty().SetOpacity(0.8)

    # The default interactor style doesn't work well for a mouse,
    # so we will redefine it as another pre-set style called 'track ball'.
    trackBallStyle = vtk.vtkInteractorStyleTrackballCamera()
    # Set the interactor style as the 'track ball' style defined above.
    iren.SetInteractorStyle(trackBallStyle)

    # This allows the interactor to initalize itself. It has to be
    # called before an event loop.
    iren.Initialize()

    # Start the event loop.
    iren.Start()
    return

def assignColors(polydata, nodeColors):
    """
    Assign colors to nodes in the given polydata based on the given colorLabels.
    ..NOTE:: The polydata is updated by reference. This function does not return any values.
    :param polydata: vtkPolyData, The polydata that is having colors assigned to it.
    :param nodeColors: array nx3, The colors that is assigned to the vertices.
    :return:
    """
    # Color the cellIds
    colors = vtk.vtkUnsignedCharArray()
    colors.SetNumberOfComponents(3)
    colors.SetName("Colors")
    for i in range(len(nodeColors)):
        nodeColor = nodeColors[i]*255
        colors.InsertNextTuple3(nodeColor[0], nodeColor[1], nodeColor[2]) # InsertNextTuple3 for (R,B,G) colours, InsertNextTuple4 to account for (R,G,B, opacity)
    polydata.GetPointData().SetScalars(colors)
    return


def getElements(triangles):
    """
    Populate a vtkCellArray with the given triangles.
    :param triangles: array nx3, An array of integers. These define triangles that are being defined
    :return: vtkCellArray
    """
    elements = vtk.vtkCellArray()
    for i in range(len(triangles)):
        # Create the object that stores the list of nodeIDs
        elementIdList = vtk.vtkIdList()
        # Insert the nodeIDs in the appropriate order.
        elementIdList.InsertNextId(triangles[i, 0])
        elementIdList.InsertNextId(triangles[i, 1])
        elementIdList.InsertNextId(triangles[i, 2])
        # Append the given vtkIdList to the end of the 'elements' vtkCellArray
        elements.InsertNextCell(elementIdList)
    return elements


def getPoints(coordinates):
    """
    Populate a vtkPoints object with the given coordinates.
    It is assumed that the indices relate to the pointId
    :param coordinates: array nx3, The coordinates of the points.
    :return: vtkPoints
    """
    nodes = vtk.vtkPoints()
    for i in range(len(coordinates)):
        nodes.InsertPoint(i, coordinates[i])
    return nodes


def getRmsError(modelresults, expresults):
    error = (modelresults - expresults)**2
    meanError = np.mean(error, axis=0)
    rmsError = np.sqrt(meanError)
    return rmsError

def getTFRdata(fileName):

    """
    reads/loads in TFRecord data format (which is of a dictionary of tensor flow arrays).
    :param fileName:
    :return:
    """

    tfFile = tensorflow.data.TFRecordDataset(fileName)

    for raw_record in tfFile.take(1):
       data = tensorflow.train.Example()
       data.ParseFromString(raw_record.numpy())
       #print(data)

    vertices = tensorflow.io.parse_tensor(data.features.feature['vertices'].bytes_list.value[0], tensorflow.float32).numpy()
    triangles = tensorflow.io.parse_tensor(data.features.feature['triangles'].bytes_list.value[0], tensorflow.int32).numpy()
    labels = tensorflow.io.parse_tensor(data.features.feature['labels'].bytes_list.value[0], tensorflow.int32).numpy()
    colors = tensorflow.io.parse_tensor(data.features.feature['vcolor'].bytes_list.value[0], tensorflow.int32).numpy()

    return data, vertices, triangles, labels, colors

if __name__ == '__main__':
    fname = "predicted_DU07_RFMB_vertx2188_aligned.dat"
    mfname = os.path.splitext(fname)[0]+"_mesh_data"+os.path.splitext(fname)[-1]

    with open(mfname, mode='rb') as fl:
        mesh_data = pickle.load(fl)

    vertices = mesh_data['vertices']
    triangles = mesh_data['faces']
    colors = mesh_data['vertex_colors']

    # color array currently showing predicted site, alter to also show actual site from labels.
    # append a column of opacity to colour matrix based on label.
    #opacityColumn = mesh_data
    #colors = np.hstack((colors,mesh_data['predictions']))

    visualiseMesh(vertices, triangles, colors)

    with open(fname, mode='rb') as fl:
        # Contains: labels, num triangles, num vertices, predictions (predicted labels), triangles, vertices.
        data = pickle.load(fl)

    acl_nodes = data['vertices'][np.where(data['labels'] != 1)] # acl node co-ordinates
    predicted_nodes = data['vertices'][np.where(data['predictions'] != 1)] # predicted acl node co-ordinates

    acl_hull = ConvexHull(acl_nodes)
    acl_centroid = np.mean(acl_nodes[acl_hull.vertices, :], axis=0)
    predict_hull = ConvexHull(predicted_nodes)
    predict_centroid = np.mean(predicted_nodes[predict_hull.vertices, :], axis=0)

    # finding the overlapping nodes
    overlap = []
    for a in acl_nodes:
        if a in predicted_nodes:
            overlap.append(a)
    overlap = np.array(overlap).reshape((-1, 3)) # convert to array and reshape
    if np.shape(overlap)[0] >= 4:
        overlap_hull = ConvexHull(overlap)
        overlap_centroid = np.mean(overlap[overlap_hull.vertices, :], axis=0)
    elif (np.shape(overlap)[0] < 4) and (np.shape(overlap)[0] > 2): #3 small triangle overlay
        overlap = np.vstack((overlap, np.mean(overlap, axis=0))).reshape((-1, 3)) # hack: add a center because ConvexHull can't take less than 4 elements
        overlap_hull = ConvexHull(overlap)
        overlap_centroid = np.mean(overlap[overlap_hull.vertices, :], axis=0)
    elif np.shape(overlap)[0] == 2: # this is a line
        overlap_hull = overlap
        overlap_centroid = np.mean(overlap, axis=0) # average between two points
    else: # there is no overlap at all
        overlap_hull = 0
        overlap_centroid = 0

    #print display info
    # Check that if 'predictions' and labels match
    print('=============================')
    print(fname)
    print('=============================')
    print('ACL Ligament Labels: ', np.where(data['labels'] == 2))
    print('Predicted ACL Labels: ', np.where(data['predictions'] == 2))
    print('RMS residual ACL error: ', np.sqrt(np.mean(np.square(predict_centroid - acl_centroid))))
    print('=============================')
    print('INFO:')
    print('Defined ACL Area: ', acl_hull.area)
    print('Defined ACL centroid: ', acl_centroid)
    print('Predicted ACL Area: ', predict_hull.area)
    print('Predicted ACL centroid: ', predict_centroid)
    print('Overlapping Area: ', overlap_hull.area)
    print('Overlapping centroid: ', overlap_centroid)

    # plotting
    fig = plt.figure()
    ax = fig.add_subplot(111, projection="3d")
    ax.plot(acl_nodes[:, 0], acl_nodes[:, 1], acl_nodes[:, 2], 'ko')
    ax.plot(predicted_nodes[:, 0], predicted_nodes[:, 1], predicted_nodes[:, 2], 'bo')
    ax.scatter(acl_centroid[0], acl_centroid[1], acl_centroid[2], s=100, c='k', marker='x')
    ax.scatter(predict_centroid[0], predict_centroid[1], predict_centroid[2], s=100, c='b', marker='x')
    ax.scatter(overlap_centroid[0], overlap_centroid[1], overlap_centroid[2], s= 100, c='r', marker='x')

    for s in overlap_hull.simplices:
        s = np.append(s, s[0])
        ax.plot(overlap[s, 0], overlap[s, 1], overlap[s, 2], "r-")

    for i in ["x", "y", "z"]:
        eval("ax.set_{:s}label('{:s}')".format(i, i))
    plt.title('Predicted vs Actual ACL Insertion Site Overlay')
    plt.show()
    fig.savefig("output_data/"+fname[:-3]+'.png')

    # write out
    output_file_name = "output_data/"+fname[:-4]+'.txt'
    with open(output_file_name,mode='w') as f:
        f.write("For test case: "+fname+'\n')
        f.write('\n')
        f.write("==== ACL predicted site information for subject ==== \n")
        f.write('\n')
        f.write("Defined ACL Area: "+str(acl_hull.area)+'\n')
        f.write("Defined ACL Centroid: " + str(acl_centroid)+'\n')
        f.write('\n')
        f.write("Predicted ACL Area: " + str(predict_hull.area)+'\n')
        f.write("Predicted ACL Centroid: " + str(predict_centroid)+'\n')
        f.write('\n')
        f.write("Overlap ACL Area: " + str(overlap_hull.area)+'\n')
        f.write("Overlap ACL Centroid: " + str(overlap_centroid)+'\n')
        f.write('\n')
        f.write('RMS residual error: ' + str(np.sqrt(np.mean(np.square(predict_centroid - acl_centroid))))+'\n')
        f.write('Absolute distance error: ' + str(np.sqrt((np.square(predict_centroid - acl_centroid)))) + '\n')
        f.write('Sum Absolute Error: ' + str(np.sqrt((np.sum(np.square(predict_centroid - acl_centroid))))) + '\n')